;-*- emacs-lisp -*- webster-19.el
;;
;; Author Benoit Grange (ben@fizz.fdn.org)
;; Modified from webster.el by Jason R.Glasgow
;;
(defvar webster-host "localhost"
  "*Host that is a webster server (use whatever convenient).")

(defvar webster-port 2627
  "*The port to connect to. Either 103 or 2627.")

(defvar webster-process nil
  "The current webster process.")

(defvar webster-buffer nil
  "The current webster buffer.")

(defvar webster-word nil
  "The current webster word.")

(defvar webster-index nil
  "The current webster index.")

(defvar webster-mode-map nil "The mode map used by webster-mode")

(defun webster-sentinel (proc change)
  "Webster process sentinel."
;;; Does nothing !!! Just avoid the messages telling about process died
)

(defvar webster-tmp-string "")

(defun webster-output (string)
  (let ((old-buffer (current-buffer)))
    (set-buffer webster-buffer)
    (goto-char (point-max))
    (cond
     ((string-match "^ERROR" string) 
      (progn 
	(insert "An error has occured.\n")
	(ding)))
     ((string-match "^SPELLING 0" string) 
      (progn
	(insert (concat "No definition for word '" webster-word "'.\n")))
      (ding))
     ((string-match "^SPELLING 1" string) 
      (insert (concat "Word '" webster-word "' is correctly spelled.\n")))
     ((string-match "^!END" string) 
      (progn 
	(insert "\n")
	(let ((webster-window (get-buffer-window webster-buffer)))
	  (if webster-window
	      (progn 
		(set-window-point webster-window webster-point)
		(message "Outputing defs...done")
		))
	  )
	))
     ((string-match "^DEFINITION 0" string)
      (progn
	(message "Outputing defs...")
	(setq webster-point (point-max))
	(insert (concat "Definition for word '" webster-word "':\n")))
      )
     ((string-match "^MATCHS" string)
      (progn
	(setq webster-point (point-max))
	(insert (concat "Endings for '" webster-word "':\n")))
      )
     (t (insert string))
     )
    (set-buffer old-buffer)
    )
)

(defun webster-filter (proc string)
  "Webster process filter."
;;; Break the string into line records
  (setq webster-tmp-string (concat webster-tmp-string string))
  (while (string-match "\n" webster-tmp-string) 
    (webster-output (substring webster-tmp-string 0 (match-end 0)))
    (setq webster-tmp-string 
	  (substring webster-tmp-string (match-end 0)))
    )
  (if (string-match "" webster-tmp-string)
      (progn
	(setq webster-tmp-string "")
	(webster-output "!END")
	)
    )
)

(defun webster-open-connection ()
  "Open a connection to the webster server, setup a new buffer if neeed."
  (interactive)
  (if (or (not webster-process)
	  (not webster-buffer)
	  (not (buffer-name webster-buffer))
	  (not (eq (process-status webster-process) 'open)))
      (progn
	(message "Connecting to webster server...")
	(setq webster-process
	      (open-network-stream "webster"
				   "*webster*" webster-host webster-port))
	(set-process-sentinel webster-process 'webster-sentinel)
	(set-process-filter webster-process 'webster-filter)
	(process-kill-without-query webster-process)
	(setq webster-buffer (process-buffer webster-process))
	(let ((old-buffer (current-buffer)))
	  (set-buffer webster-buffer)
	  (webster-mode)
	  )
	(setq webster-index 'dictionary)
	(message "Connecting to webster server...done")
	)
    )
  )

(defun webster-close-connection ()
  "Open a connection to the webster server, setup a new buffer if neeed."
  (interactive)
  (if (not (eq (process-status webster-process) 'open))
      (delete-process webster-process)
    )
  )

(defun webster-send-request (kind word)
  "Send the request to the webster server"
  (webster-open-connection)
  (process-send-string webster-process (concat kind " " word "\n"))
  (display-buffer webster-buffer)
)

(defun webster (arg)
  "Look up a word in the Webster's dictionary.
Open a network communication to a webster server specified in the webster-host variable."
  (interactive (list
		(read-string
		 (concat "Look up word in webster (" (current-word) "): "))))
  (if (equal "" arg) (setq arg (current-word)))
  (setq webster-word arg)
  (webster-send-request "DEFINE" arg)
)

(defun webster-spell (arg)
  "Check a word spelling in the Webster's dictionary.
Open a network communication to a webster server specified in the webster-host variable."
  (interactive (list
		(read-string
		 (concat "Spell up word in webster (" (current-word) "): "))))
  (if (equal "" arg) (setq arg (current-word)))
  (setq webster-word arg)
  (webster-send-request "SPELL" arg)
)

(defun webster-endings (arg)
  "Print all word endings found in the Webster's dictionary.
Open a network communication to a webster server specified in the webster-host variable."
  (interactive (list
		(read-string
		 (concat "Provide endings for word (" (current-word) "): "))))
  (if (equal "" arg) (setq arg (current-word)))
  (setq webster-word arg)
  (webster-send-request "ENDINGS" arg)
)

(defun webster-mode ()
  "Major mode for interacting with on-line Webster's dictionary.
\\{webster-mode-map}
Use webster-mode-hook for customization."
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'webster-mode)
  (setq mode-name "Webster")
  (use-local-map webster-mode-map)
  (run-hooks 'webster-mode-hook)
  (if webster-mode-map
      nil
    (setq webster-mode-map (make-sparse-keymap))
    (define-key webster-mode-map "?" 'describe-mode)
    (define-key webster-mode-map "d" 'webster)
    (define-key webster-mode-map "e" 'webster-endings)
    (define-key webster-mode-map "q" 'webster-quit)
    (define-key webster-mode-map "s" 'webster-spell))
)
(defun webster-quit ()
   "Close connection and quit webster-mode.
Buffer is not deleted."
   (interactive)
   (webster-close-connection)
   (bury-buffer webster-buffer))
